NAME = "data portal"

ID = "orangecontrib.tomwer.widgets.dataportal"

DESCRIPTION = "widgets for data portal publication"

LONG_DESCRIPTION = "Contains widget to publish data to the esrf data portal"

ICON = "../../widgets/icons/tomwer_data_portal.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 13
